﻿
--Create the addon using Ace3
BRAnkhTimer = LibStub("AceAddon-3.0"):NewAddon("Broker_AnkhTimer", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

--Create the data broker
BRAnkhTimer.dataObject = LibStub:GetLibrary("LibDataBroker-1.1"):NewDataObject("AnkhTimer", {
	type = "data source",
	text = "Ankhs",
    icon = "Interface\\AddOns\\Broker_AnkhTimer\\icon.tga",
    label = "AnkhTimer"
})


--Localization
local L = LibStub("AceLocale-3.0"):GetLocale("Broker_AnkhTimer", true)

--Local variables
local ankhsInBags = 0
local reincarnationCooldown = 0
local timerHandle = nil


--###############################################################
--AceAddon call back fuctions
--###############################################################


function BRAnkhTimer:OnInitialize()
	
	--Define configuration options
	local options = {
    	type ="group",	
		name = "AnkhTimer",
    	handler = BRAnkhTimer,
    	childGroups = "tree",
    	args = {
    		header1 = {
				type = "description",
				name = "v0.9",
				order = 1,
				width = "full",
				cmdHidden = true
			},
    		showCount = {
            	type = "toggle",
            	name = "Show count",
            	order = 2,
				width = "full",
            	desc = "Show the number of Ankhs in the bag",
           	 	set = function (info, value) 
           	 			self.db.profile.showCount = value 
           	 			self:RefreshText()
           	 		  end,
            	get = function (info) return self.db.profile.showCount end
        	},
        	showCooldown = {
            	type = "toggle",
            	name = "Show cooldown",
            	order = 3,
				width = "full",
            	desc = "Show the Ankh cooldown",
           	 	set = function (info, value) 
           	 			self.db.profile.showCooldown = value 
           	 			self:RefreshText()
           	 		  end,
            	get = function (info) return self.db.profile.showCooldown end
        	},
        	autoBuyAnkhs = {
            	type = "toggle",
            	name = "Auto-buy Ankhs",
            	order = 4,
				width = "full",
            	desc = "Automatically but Ankhs from vendors",
           	 	set = function (info, value) self.db.profile.autoBuyAnkhs = value end,
            	get = function (info) return self.db.profile.autoBuyAnkhs end
        	},
    		maxAnkhs = {
            	type = "input",
            	order = 5,
            	name = "Maximum Ankhs",
            	desc = "Number of Ankhs to try and keep in your bags",
           	 	set = "SetMaxAnkhs",
            	get = "GetMaxAnkhs"
        	},
        	testPrint = {
            	type = "execute",
            	name = "TestPrint",
            	order = 6,
            	hidden = true,
            	desc = "Print out the internal state of the addon",
           	 	func = "TestPrint"
        	},
        	simulateDeath = {
            	type = "execute",
            	name = "SimulateDeath",
            	order = 7,
            	hidden = true,
            	desc = "Simulate that the player used Reincarnation",
           	 	func = "SimulateDeath"
        	},
        	debugPrints = {
            	type = "toggle",
            	name = "Debug Prints",
            	order = 8,
            	hidden = true,
            	desc = "Enable debug prints",
           	 	set = function (info, value) self.db.profile.debugPrints = value end,
            	get = function (info) return self.db.profile.debugPrints end
            },
    	},
	}
	
	
	--Register options and set up configuration options
	local AceCfg = LibStub("AceConfig-3.0")
	local AceCfgReg = LibStub("AceConfigRegistry-3.0")
	local AceDlg = LibStub("AceConfigDialog-3.0")
	
	--Register options and slash commands
	AceCfg:RegisterOptionsTable("Broker_AnkhTimer", options, {"Broker_AnkhTimer", "BRAT"})
	
	--Set up configuration using Blizzards addon tab
	--Check if "Broker" group exists, in not create one
	local brokerOptions = AceCfgReg:GetOptionsTable("Broker", "dialog", "LibDataBroker-1.1")
	if (not brokerOptions) then
		brokerOptions = {
			type = "group",
			name = "Broker",
			args = {
			}
		}
		AceCfg:RegisterOptionsTable("Broker", brokerOptions)
		AceDlg:AddToBlizOptions("Broker", "Broker")
	end
	--Add addon configuration options
	self.optionsFrame = AceDlg:AddToBlizOptions("Broker_AnkhTimer", "AnkhTimer", "Broker")
	
	--Default database values
	local dbDefaults = {
		profile = {
			debugPrints = false,
			showCount = true,
			showCooldown = true,
			autoBuyAnkhs = true,
			maxAnkhs = 10,
			reincarnationLastUsed = 0
		}
	}	
		
	--Set up the database
	self.db = LibStub("AceDB-3.0"):New("Broker_AnkhTimerDB", dbDefaults)
	
	self:DebugPrint("OnInitialize")
	
end


function BRAnkhTimer:OnEnable()
	self:DebugPrint("OnEnable")
	
	--Calculate start values
	ankhsInBags = self:NumberOfAnkhsInBags()
	reincarnationCooldown = self:getReincarnationCooldown()
	
	--Check if we need to start the cooldown timer
	if ((time() - self.db.profile.reincarnationLastUsed) < reincarnationCooldown ) then
		self:DebugPrint("Still on cooldown, start timer")
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
		--No need to call UpdateLDB here, the timer will do that for us
	elseif (self:isReincarnationReady() == false) then
		--Special case, the spell is on cooldown but not according to the time stored in self.db.profile.reincarnationLastUsed
		--This can happen the first time the addon is installed or if the spell is used without the addon present (addon disabled, another computer etc)
		self.db.profile.reincarnationLastUsed = 0
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
	end
	
	--Checking when spell and talent changes, see if the Reincarnat cooldown changes
	self:RegisterEvent("SPELLS_CHANGED", "TalentCheck")
    self:RegisterEvent("LEARNED_SPELL_IN_TAB", "TalentCheck")
    
    --Detect when player dies and reserect
    self:RegisterEvent("PLAYER_DEAD", "PlayerDead")
    
    --Detect when a player zone in and out, used to register for BAG_UPDATE events
    self:RegisterEvent("PLAYER_ENTERING_WORLD", "PlayerEnteringWorld")
    self:RegisterEvent("PLAYER_LEAVING_WORLD", "PlayerLeavingWorld")
    
    --Detect when trading with a vendor to auto-buy Ankhs
    self:RegisterEvent("MERCHANT_SHOW", "MerchantShow")

end


function BRAnkhTimer:OnDisable()
    self:DebugPrint("OnDisable")
    
    --Unregister all events
    self:UnregisterAllEvents()
    
    --Stop the timer if is running
    if (timerHandle ~= nil) then
		--Turn off the timer
		self:DebugPrint("Turn off the timer")
		self:CancelTimer(timerHandle)
		timerHandle = nil
    end
    
    --Update the data object
	self:RefreshText()
    
end


--###############################################################
--Event callbacks
--###############################################################

--Check if the cooldown of the Reincarnation spell has changed
function BRAnkhTimer:TalentCheck()
	self:DebugPrint("TalentCheck")
	
	--Get the cooldown
	local cooldown = self:getReincarnationCooldown()
	
	--check if the cooldown has changed
	if (cooldown ~= reincarnationCooldown) then
		self:DebugPrint("Reincarnation cooldown time changed")
		reincarnationCooldown = cooldown
		
		--Update the data object
		self:RefreshText()
		
	end
	
end


--Player is alive, check if it is due to using the Reincarnation spell
function BRAnkhTimer:PlayerAlive()
	self:DebugPrint("PlayerAlive")
	
	--Check if player alive is due to using the Reincarnation spell
	if (self:isReincarnationReady() == false) then
		self:DebugPrint("Player used Ankh to reserect")
		
		--Store the time when Ankh was used
		self.db.profile.reincarnationLastUsed = time()
		
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
		--No need to call UpdateLDB here, the timer will do that for us
		
	else
		self:DebugPrint("Ankh was not used")
	end
	
	--Stop listening to the PLAYER_ALIVE event
	self:UnregisterEvent("PLAYER_ALIVE")
	
	--Unfortenatly I can not check the start time using the Blizz function start, duration, enable = GetSpellCooldown(spellName, "BOOKTYPE") here
	--The PLAYER_ALIVE message probably happens before the new values are set for the spell
	--Start a timer event and check the values 1 seconds after the PLAYER_ALIVE message
	timerHandle = self:ScheduleTimer("PlayerAliveTimerEvent", 1, time())
	
end


--Called after PLAYER_ALIVE event with a short delay need to allow the game set spell values
function BRAnkhTimer:PlayerAliveTimerEvent(reserectionTime)
	--Check if player alive is due to using the Reincarnation spell
	if (self:isReincarnationReady() == false) then
		self:DebugPrint("2:Player used Ankh to reserect")
		
		--Store the time when Ankh was used
		self.db.profile.reincarnationLastUsed = reserectionTime
		
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
		--No need to call UpdateLDB here, the timer will do that for us
		
	else
		self:DebugPrint("2:Ankh was not used")
	end

end


--Player is dead
function BRAnkhTimer:PlayerDead()
	self:DebugPrint("PlayerDead")
	
	--Check if reincarnation is available and ready to be used, then the next PLAYER_ALIVE event might by due to player using the reincarnation spell
	if (self:isReincarnationReady() == true) then
		self:RegisterEvent("PLAYER_ALIVE", "PlayerAlive")
	end
	
end


--Player entering world, register for BAG_UPDATE events
function BRAnkhTimer:PlayerEnteringWorld()
	self:DebugPrint("PlayerEnteringWorld")
	
	self:RegisterEvent("BAG_UPDATE", "BagUpdate")
end


--Player leaving world, unregister for BAG_UPDATE events to keep load down
function BRAnkhTimer:PlayerLeavingWorld()
	self:DebugPrint("PlayerLeavingWorld")
	
	self:UnregisterEvent("BAG_UPDATE")
end


--Bags are updated, check if the number of Ankhs has changed
function BRAnkhTimer:BagUpdate()
	self:DebugPrint("BagUpdate")
	
	local ankhs = self:NumberOfAnkhsInBags()
	--check if it differs for before the event
	if (ankhs ~= ankhsInBags) then
		self:DebugPrint("Number of Ankhs in the bag changed")
		ankhsInBags = ankhs
		
	--Update the data object
	self:RefreshText()
		
	end

end


--Trading with a vendor, try to auto-buy Ankhs
function BRAnkhTimer:MerchantShow()
	self:DebugPrint("MerchantShow")

	--Check if the current player is a shaman, if not never auto-buy anything	
	if select(2, UnitClass("player")) ~= "SHAMAN" then
		return
	end

	if (self.db.profile.autoBuyAnkhs == true) then
	
		--Get the number of items the vendor has
		local numItems = GetMerchantNumItems()
		
		--Check each item
		for index = 1, numItems, 1 do
			-- Get the item link
			local itemLink = GetMerchantItemLink(index)
		
			if (itemLink ~= nil) then
				--Extract the itemstring
		    	local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
		    
		        --split up the item sting to find the itemID
		    	local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemString)
		    	
		    	--Check if its an Ankh
		    	if (itemId == "17030") then
		    		self:DebugPrint("Vendor is selling Ankhs")
		    		--Auto-buy if needed
		    		if (ankhsInBags < self.db.profile.maxAnkhs) then
		    			self:DebugPrint("By a few Ankhs from the vendor")
		    			BuyMerchantItem(index , (self.db.profile.maxAnkhs - ankhsInBags))
		    			
		    			--Print a message, its nice if the use know what addon is buying stuff
		    			DEFAULT_CHAT_FRAME:AddMessage("|cffffff00AnkhTimer addon is restocking your Ankhs.|r")
		    			
		    		end
		    	end
		    	
			end
			
		end
		
	end	
	
end


--Called once per second when the reincarnation spell is on cooldown
function BRAnkhTimer:CooldownTimerEvent()
	--Check if the timer should be turned off
	if (self:isReincarnationReady() == true) then
		--Turn off the timer
		self:DebugPrint("Turn off the timer")
		self:CancelTimer(timerHandle)
		timerHandle = nil
	end
	
    --Update the data object
	self:RefreshText()

end



--###############################################################
--Data Broker methods
--###############################################################

--Update the data object with new text
function BRAnkhTimer:RefreshText()
	--self:DebugPrint("UpdateDataObject")
	
    --Update the data object
    self.dataObject.text = self:getDisplayString()
    
    --Inform listeners that they should refesh them self
    --self:Fire("Broker_AnkhTimer_Refresh")
   
end


--Create the text string to display
function BRAnkhTimer:getDisplayString()
    local s = ""
    
    --Create the count part of the string
    if (self.db.profile.showCount == true) then
    	s = s .. format("|cff%s%d|r Ankh(s)", self:GetHexColor(ankhsInBags/10), ankhsInBags)
    	--self:DebugPrint("Data object ankh text = " .. s)
    end
    
    --Create the cooldown part of the string
    if (self.db.profile.showCooldown == true) then
    	local deltaTime = time() - self.db.profile.reincarnationLastUsed
    	--Check if the delta time is less the the reincarnation cooldown time
    	if (deltaTime < reincarnationCooldown) then
    		--Show current cooldown
    		local remainingTime = reincarnationCooldown - deltaTime
    		s = s .. format(" |cff%s%s|r", self:GetHexColor(1-remainingTime/reincarnationCooldown), date("%M:%S", remainingTime))
    		--self:DebugPrint("Data object ankh + cooldown text = " .. s)
    	elseif (self:isReincarnationReady() == false) then
    		--Special case, spell is on cooldown but not possible to calculate
    		s = s .. format(" |cffffff00Unkown|r")
    	end
    end
    
    return s
    
end   


--Called from LDB when the tooltip should be updated
function BRAnkhTimer.dataObject.OnTooltipShow(tooltip)

	local reincarnationLastUsed = BRAnkhTimer.db.profile.reincarnationLastUsed
	--Color codes
	local whiteColor = "|cffffffff%s|r"
	
	tooltip:AddLine("AnkhTimer")
 	--The date Reincarnation was used the last time
    if (reincarnationLastUsed ~= 0) then
  		tooltip:AddDoubleLine("Reincarnated since: ", format(whiteColor, date("%c", reincarnationLastUsed)))
    else
    	tooltip:AddDoubleLine("Reincarnated since:", format(whiteColor, "Unknown"))
    end
    	 
    --The number of Ankhs in the bags
    tooltip:AddDoubleLine("Ankhs: ", format(whiteColor, ankhsInBags))

    --The number of minutes until Ankh can be used again
    local deltaTime = time() - reincarnationLastUsed
    if( deltaTime < reincarnationCooldown ) then
    	local remainingTime = reincarnationCooldown - deltaTime
        tooltip:AddDoubleLine("Cooldown: ", format(whiteColor, date("%M:%S", remainingTime)))
    elseif (BRAnkhTimer:isReincarnationReady() == false) then
    	--Special case, Spell is on cooldown but not possible to caculate
    	tooltip:AddDoubleLine("Cooldown:", format(whiteColor, "Cooldown unknown"))
    else
        tooltip:AddDoubleLine("Cooldown:", format(whiteColor, "Ankh is ready"))
    end
    
    --The reincrnation spell cooldown
    tooltip:AddDoubleLine("Ability cooldown: ", format(whiteColor, ((reincarnationCooldown/60).." min")))
    
    tooltip:AddLine("|cffffff00Right-click|r to open options menu")
 
end


--Called when klicking on the addon
function BRAnkhTimer.dataObject.OnClick(self, button)
	--Open the configuraton menu if right button was used
	if (button == "RightButton") then
		InterfaceOptionsFrame_OpenToCategory(BRAnkhTimer.optionsFrame)
	end
end




--###############################################################
--Debug prints
--###############################################################

function BRAnkhTimer:DebugPrint(text)
	--check if debug is enabled
	if (text ~= nil and self.db.profile.debugPrints == true) then
		self:Print(text)
	end
end



--###############################################################
--Configuration callback methods
--###############################################################


function BRAnkhTimer:SetMaxAnkhs(info, value) 
	self:DebugPrint("SetMaxAnkhs")

	if (value ~=nil) then
		--Try and convert to a number
		num = tonumber(value)
		if (num ~=nil) then
			self.db.profile.maxAnkhs = num 
		else
			self:Print("Must be a number")
		end
	end
 
end
            	

function BRAnkhTimer:GetMaxAnkhs(info) 
	self:DebugPrint("GetMaxAnkhs")
	local str = tostring(self.db.profile.maxAnkhs)
	if (str ~= nil) then
		return str
	else
		return ""
	end

end


function BRAnkhTimer:TestPrint()
	self:Print("Printing out internal state of the addon")
	
	self:Print("Local variables")
	self:Print("--ankhsInBags = " .. ankhsInBags)
	self:Print("--reincarnationCooldown = " .. reincarnationCooldown)

	if (timerHandle ~= nil) then
		self:Print("--timerHandle = " .. timerHandle)
	else
		self:Print("--timerHandle = nil")
	end
	
	self:Print("DB settings")
	self:Print("--debugPrints = " .. self:BooleanToString(self.db.profile.debugPrints))
	self:Print("--showCount = " .. self:BooleanToString(self.db.profile.showCount))
	self:Print("--showCooldown = " .. self:BooleanToString(self.db.profile.showCooldown))
	self:Print("--autoBuy = " .. self:BooleanToString(self.db.profile.autoBuyAnkhs))
	self:Print("--maxAnkhs = " .. self.db.profile.maxAnkhs)
	self:Print("--reincarnationLastUsed = " .. self.db.profile.reincarnationLastUsed)
	
	self:Print("Calculated values")
	self:Print("--ankhsInBags = " .. self:NumberOfAnkhsInBags())
	self:Print("--reincarnationCooldown = " .. self:getReincarnationCooldown())
	self:Print("--isReincarnationReady = " .. self:BooleanToString(self:isReincarnationReady()))

end


--Simple help function for printing out booleans
function BRAnkhTimer:BooleanToString(value)
	if (value == nil) then 
		return "nil"
	elseif (value == true) then
		return "true"
	else
		return "false"
	end
end


--Simulate the the player died and used reincarnation spell
function BRAnkhTimer:SimulateDeath()
	self:DebugPrint("SimulateDeath")

	--Only possible if reincarnation is not on cooldown
	if (timerHandler == nil) then
		--Set last timer used to 1 minute ago
		self.db.profile.reincarnationLastUsed = time() - (59*60)
		--Start timer to exectue the countdown
		timerHandle = self:ScheduleRepeatingTimer("CooldownTimerEvent", 1)
		--No need to can UpdateLDB here, the timer will do that for us
	else
		self:Print("Not possible to simulate, cooldown already running")
	end

end




--###############################################################
--Help functions
--###############################################################

--TODO make these fuctions local?????

--Count the number of Ankhs currently in the bags
function BRAnkhTimer:NumberOfAnkhsInBags()
	
	local ankhs = 0
	
	--Check all 5 bags
	for bag=0, 4, 1 do
		
		--Number of slots in the bag
        local bagSize = GetContainerNumSlots(bag)
        
        if (bagSize > 0) then
			--Check each slot in the bag
            for slot=1, bagSize, 1 do	
	            --Get the item link
                local itemLink = GetContainerItemLink(bag, slot)
           		
                --Ceck that the slot actually contains someting
            	if (itemLink ~=nil) then
            		
            	    --Extract the itemstring
                	local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
                
	                --split up the item sting to find the itemID
	            	local _, itemId, enchantId, jewelId1, jewelId2, jewelId3, jewelId4, suffixId, uniqueId = strsplit(":", itemString)
	            	
	            	--Check if its an Ankh
	            	if (itemId == "17030") then
	            		local texture, itemCount = GetContainerItemInfo(bag, slot)
	            		ankhs  = ankhs + itemCount
	            	end	
            	end 
   			end
   		end
    
   	end
    
    return ankhs
    
end


--Get the Reincarnation cooldown based in the current talents set
function BRAnkhTimer:getReincarnationCooldown()

	--Find the number of talents set for Improved Self-Reincarnation
    local talentPoints = 0
    
    --Check each tab
    for talent = 1, GetNumTalentTabs(), 1 do
    	--Check each talent
        for talentIdx  = 1, GetNumTalents(talent), 1 do
            local name, icon, iconx, icony, currRank, maxRank = GetTalentInfo(talent, talentIdx)
            
            if (name ~= nil) then
                if (name == "Improved Reincarnation") then
                    talentPoints = currRank
                end
            end
            
        end
    end
	
    --Convert to seconds    
    return (3600 - (10 * 60 * talentPoints))

end


--Check if the reincarnation spell is ready do be used
function BRAnkhTimer:isReincarnationReady()
	local spellName = GetSpellInfo(20608)
	local start, duration, enabled = GetSpellCooldown(spellName, "BOOKTYPE")
	if (start == 0) then
		return true
	else
		return false
	end
end


--Return a color code base on the the percentage figure
function BRAnkhTimer:GetHexColor(p)
    if p > 0.8 then
    	--Green
    	return "00ff00"
    elseif p > 0.6 then
    	--Yellow
    	return "ffff00"
    elseif p > 0.4 then
    	--Orange
     	return "ff7f00"
    elseif p > 0.2 then
    	--xxx
      return "ff7f00"
    else
    	--Red
    	return "ff0000"
    end
end

--TODO:
--1. Localization


--Suggestions:
--Dialog warning when low on Ankhs
--Review the colors used
--Use colors on the tooptip
--Dynamic tooltip to shown countdown of cooldown
--Notification when the coodown is up (what kind of notification?)

--Questions:
--Any troubles when another class run this addon by misstake?
--Test with other data broker displays
--Any issues when low-level shamans are using this addon?
--Some special cases during and after PLAYER_ENTERING_WORLD when the number of slots in the bags are set to 0, some cashing issues
--Are database values stored unique for each charcter?

--Limitations
--Not possible to calcualte and show the cooldown if on cooldown when the addon is ussed
--If reincarnation is ued with the addon not pressent (disabled, on another computer etc) its not possible to calculate and show the cooldown

--Knowbugs:
--

